/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2022 Loongson Technology Corporation Limited
 */


#ifndef _L3FWD_MIPS_H_
#define _L3FWD_MIPS_H_

#include "l3fwd.h"
#include "l3fwd_common.h"


/*
 * Update source and destination MAC addresses in the ethernet header.
 * Perform RFC1812 checks and updates for IPV4 packets.
 */
static inline void
processx4_step3(struct rte_mbuf *pkt[FWDSTEP], uint16_t dst_port[FWDSTEP])
{
#ifdef RTE_ARCH_NO_VECTOR
	xmm_t te[FWDSTEP];
	xmm_t te2[FWDSTEP];
	xmm_t ve[FWDSTEP];
	xmm_t *p[FWDSTEP];

	p[0] = rte_pktmbuf_mtod(pkt[0], xmm_t *);
	p[1] = rte_pktmbuf_mtod(pkt[1], xmm_t *);
	p[2] = rte_pktmbuf_mtod(pkt[2], xmm_t *);
	p[3] = rte_pktmbuf_mtod(pkt[3], xmm_t *);

	ve[0] = val_eth[dst_port[0]];

	ve[1] = val_eth[dst_port[1]];

	ve[2] = val_eth[dst_port[2]];

	ve[3] = val_eth[dst_port[3]];

	te[0] = *p[0];
	te[1] = *p[1];
	te[2] = *p[2];
	te[3] = *p[3];

	/* Update first 12 bytes, keep rest bytes intact. */
	//r0 := (mask0 == 0) ? a0 : b0
	/* MASK_ETH = 0x3f */

	te2[0] = ve[0];
	te2[0].i32[3] = te[0].i32[3];
	te2[1] = ve[1];
	te2[1].i32[3] = te[1].i32[3];
	te2[2] = ve[2];
	te2[2].i32[3] = te[2].i32[3];
	te2[3] = ve[3];
	te2[3].i32[3] = te[3].i32[3];

	*p[0] = te2[0];
	*p[1] = te2[1];
	*p[2] = te2[2];
	*p[3] = te2[3];

	rfc1812_process((struct ipv4_hdr *)
			((struct ether_hdr *)p[0] + 1),
			&dst_port[0], pkt[0]->packet_type);
	rfc1812_process((struct ipv4_hdr *)
			((struct ether_hdr *)p[1] + 1),
			&dst_port[1], pkt[1]->packet_type);
	rfc1812_process((struct ipv4_hdr *)
			((struct ether_hdr *)p[2] + 1),
			&dst_port[2], pkt[2]->packet_type);
	rfc1812_process((struct ipv4_hdr *)
			((struct ether_hdr *)p[3] + 1),
			&dst_port[3], pkt[3]->packet_type);

#endif
}

/*
 * Group consecutive packets with the same destination port in bursts of 4.
 * Suppose we have array of destionation ports:
 * dst_port[] = {a, b, c, d,, e, ... }
 * dp1 should contain: <a, b, c, d>, dp2: <b, c, d, e>.
 * We doing 4 comparisons at once and the result is 4 bit mask.
 * This mask is used as an index into prebuild array of pnum values.
 */
static inline uint16_t *
port_groupx4(uint16_t pn[FWDSTEP + 1], uint16_t *lp, xmm_t dp1, xmm_t dp2)
{
	union {
		uint16_t u16[FWDSTEP + 1];
		uint64_t u64;
	} *pnum = (void *)pn;

	int32_t v = 0;
	uint32_t i;
	xmm_t tmp, tmp1;

#ifdef RTE_ARCH_NO_VECTOR
	for (i = 0; i < 8; i++)
		dp1.i16[i] = (dp1.i16[i] == dp2.i16[i]) ? 0xffff : 0;
	tmp = tmp1 = dp1;
	for (i = 0; i < 4; i++) {
		dp1.i16[2 * i] = tmp.i16[i];
		dp1.i16[2 * i + 1] = tmp1.i16[i];
	}
#endif

	for (i = 0; i < 4; i++)
		if (dp1.i32[i] < 0)
			v |= (1 << i);

	/* update last port counter. */
	lp[0] += gptbl[v].lpv;

	/* if dest port value has changed. */
	if (v != GRPMSK) {
		pnum->u64 = gptbl[v].pnum;
		pnum->u16[FWDSTEP] = 1;
		lp = pnum->u16 + gptbl[v].idx;
	}

	return lp;
}

/**
 * Process one packet:
 * Update source and destination MAC addresses in the ethernet header.
 * Perform RFC1812 checks and updates for IPV4 packets.
 */
static inline void
process_packet(struct rte_mbuf *pkt, uint16_t *dst_port)
{
#ifdef RTE_ARCH_NO_VECTOR
	struct ether_hdr *eth_hdr;
	xmm_t te, ve;

	eth_hdr = rte_pktmbuf_mtod(pkt, struct ether_hdr *);

	te = *(xmm_t *)eth_hdr;

	ve = val_eth[dst_port[0]];

	rfc1812_process((struct ipv4_hdr *)(eth_hdr + 1), dst_port,
			pkt->packet_type);

	te.i32[0] = ve.i32[0];
	te.i32[1] = ve.i32[1];
	te.i32[2] = ve.i32[2];
	*(xmm_t *)eth_hdr = te;

#endif
}

/**
 * Send packets burst from pkts_burst to the ports in dst_port array
 */
static __rte_always_inline void
send_packets_multi(struct lcore_conf *qconf, struct rte_mbuf **pkts_burst,
		uint16_t dst_port[MAX_PKT_BURST], int nb_rx)
{
#ifdef RTE_ARCH_NO_VECTOR
	int32_t k;
	int j = 0;
	uint16_t dlp;
	uint16_t *lp;
	uint16_t pnum[MAX_PKT_BURST + 1];
	uint16_t tmp;
	uint64_t tmp_hi, tmp_lo;

	/*
	 * Finish packet processing and group consecutive
	 * packets with the same destination port.
	 */
	k = RTE_ALIGN_FLOOR(nb_rx, FWDSTEP);
	if (k != 0) {
		xmm_t dp1, dp2;

		lp = pnum;
		lp[0] = 1;

		processx4_step3(pkts_burst, dst_port);

		/* dp1: <d[0], d[1], d[2], d[3], ... > */
		dp1 = *(xmm_t *)dst_port;
		tmp = (FWDSTEP - 1) * sizeof(dst_port[0]);

		for (j = FWDSTEP; j != k; j += FWDSTEP) {
			processx4_step3(&pkts_burst[j], &dst_port[j]);

			/*
			 * dp2:
			 * <d[j-3], d[j-2], d[j-1], d[j], ... >
			 */
			dp2 = *(xmm_t *)&dst_port[j - FWDSTEP + 1];
			lp  = port_groupx4(&pnum[j - FWDSTEP], lp, dp1, dp2);

			/*
			 * dp1:
			 * <d[j], d[j+1], d[j+2], d[j+3], ... >
			 */
			if (tmp > 15)
				tmp = 16;
			tmp *= 8;
			tmp_hi = dp2.i64[1];
			tmp_lo = dp2.i64[0];
			if (tmp > 64) {
				tmp_lo = (tmp_hi >> (tmp - 64));
				tmp_hi = 0;
			} else {
				tmp_lo = (tmp_lo >> tmp) | (tmp_hi << (64 - tmp));
				tmp_hi = tmp_hi >> tmp;
			}
			dp1.i64[1] = tmp_hi;
			dp1.i64[0] = tmp_lo;
		}

		/*
		 * dp2: <d[j-3], d[j-2], d[j-1], d[j-1], ... >
		 */
		/* 0xf9 = 0x11..0x11..0x10..0x01 */
		dp2.i64[1] = dp1.i64[1];

		dp2.i16[0] = dp1.i16[1];
		dp2.i16[1] = dp1.i16[2];
		dp2.i16[2] = dp1.i16[3];
		dp2.i16[3] = dp1.i16[3];
		lp  = port_groupx4(&pnum[j - FWDSTEP], lp, dp1, dp2);

		/*
		 * remove values added by the last repeated
		 * dst port.
		 */
		lp[0]--;
		dlp = dst_port[j - 1];
	} else {
		/* set dlp and lp to the never used values. */
		dlp = BAD_PORT - 1;
		lp = pnum + MAX_PKT_BURST;
	}

	/* Process up to last 3 packets one by one. */
	switch (nb_rx % FWDSTEP) {
	case 3:
		process_packet(pkts_burst[j], dst_port + j);
		GROUP_PORT_STEP(dlp, dst_port, lp, pnum, j);
		j++;
		/* fall-through */
	case 2:
		process_packet(pkts_burst[j], dst_port + j);
		GROUP_PORT_STEP(dlp, dst_port, lp, pnum, j);
		j++;
		/* fall-through */
	case 1:
		process_packet(pkts_burst[j], dst_port + j);
		GROUP_PORT_STEP(dlp, dst_port, lp, pnum, j);
		j++;
	}

	/*
	 * Send packets out, through destination port.
	 * Consecutive packets with the same destination port
	 * are already grouped together.
	 * If destination port for the packet equals BAD_PORT,
	 * then free the packet without sending it out.
	 */
	for (j = 0; j < nb_rx; j += k) {

		int32_t m;
		uint16_t pn;

		pn = dst_port[j];
		k = pnum[j];

		if (likely(pn != BAD_PORT))
			send_packetsx4(qconf, pn, pkts_burst + j, k);
		else
			for (m = j; m != j + k; m++)
				rte_pktmbuf_free(pkts_burst[m]);

	}

#endif
}

#endif /* _L3FWD_MIPS_H_ */
